classdef EEGBrowser < CoreBrowser
    properties
        windowWidth = 5;
        normalizeFlag = false;
        showChannelNumber = false;
        gain = 0.5;
        numberOfChannelsToPlot
        yTickLabel
        colormap = 'eegplot';
        colorInCell
        textHandle;
        isEpoched
        dim
    end
    properties(SetObservable)
        color
    end
    methods
        %% constructor
        function obj = EEGBrowser(EEG, objMaster)
            if nargin < 2, objMaster = -1;end
            obj.dim = size(EEG.data);
            obj.isEpoched = false;
            if length(obj.dim) > 2
                obj.isEpoched = true;
            end
            if obj.isEpoched
                EEG.data = reshape(EEG.data,[obj.dim(1) prod(obj.dim(2:3))]);
            end
            ntimePoints = size(EEG.data,2);
            obj.streamHandle.name = EEG.setname;
            obj.streamHandle.timeStamp = (0:ntimePoints-1)/EEG.srate;
            obj.streamHandle.numberOfChannels = obj.dim(1);
            obj.streamHandle.mmfName = [tempdir filesep obj.streamHandle.name '.bin'];
            precision = class(EEG.data(1));
            fid = fopen(obj.streamHandle.mmfName,'w');
            delta = ceil(0.2*obj.streamHandle.numberOfChannels);
            for it=1:delta:obj.streamHandle.numberOfChannels
                if it+delta-1 <=obj.streamHandle.numberOfChannels
                    fwrite(fid,EEG.data(it:it+delta-1,:)',precision);
                else
                    fwrite(fid,EEG.data(it:end,:)',precision);
                end
            end
            fclose(fid);
            obj.streamHandle.mmfObj = memmapfile(obj.streamHandle.mmfName,'Format',{precision [length(obj.streamHandle.timeStamp) obj.streamHandle.numberOfChannels] 'x'},'Writable',false);
            % obj.streamHandle.data = EEG.data(:,1:end)';
            obj.streamHandle.event = event;
            obj.streamHandle.samplingRate = EEG.srate;
            try
                obj.streamHandle.label = {EEG.chanlocs.labels};
            catch 
                for it=1:obj.streamHandle.numberOfChannels, obj.streamHandle.label{it} = num2str(it);end
            end
            while length(obj.streamHandle.label) < obj.streamHandle.numberOfChannels
                obj.streamHandle.label{end+1} = num2str(length(obj.streamHandle.label)+1);
            end
            obj.streamHandle.chanlocs = EEG.chanlocs;
            obj.streamHandle.icawinv = EEG.icawinv;
            obj.streamHandle.icachansind = EEG.icachansind;
            if isfield(EEG.event,'type')
                type = {EEG.event.type};
                for ev=1:length(type), if isnumeric(type{ev}), type{ev} = num2str(type{ev});end;end
                latency = round(cell2mat({EEG.event.latency}));
                if obj.isEpoched
                    latency = [latency [1:obj.dim(2):prod(obj.dim(2:3)) prod(obj.dim(2:3))]];
                    type = [type repmat({''},1,obj.dim(3)+1)];
                end
                obj.streamHandle.event = obj.streamHandle.event.addEvent(latency,type);
            end
            obj.addlistener('channelIndex','PostSet',@EEGBrowser.updateChannelDependencies);
            obj.addlistener('timeIndex','PostSet',@EEGBrowser.updateTimeIndexDenpendencies);
            obj.addlistener('color','PostSet',@EEGBrowser.updateColorInCell);
            obj.master = objMaster;
            obj.init();
        end
        %% plot
        function init(obj)
            init@CoreBrowser(obj);
            set(obj.figureHandle,'Renderer','Painters','CloseRequestFcn',@(src, event)onClose(obj,[], event))
            set(obj.figureHandle,'name','Scroll activities -- pop_eegbrowser()');
            
            % find now cursor index
            [~,t1] = min(abs(obj.streamHandle.timeStamp(obj.timeIndex) - (obj.nowCursor-obj.windowWidth/2)));  
            [~,t2] = min(abs(obj.streamHandle.timeStamp(obj.timeIndex) - (obj.nowCursor+obj.windowWidth/2)));  
            data = obj.streamHandle.mmfObj.Data.x(obj.timeIndex(t1:t2),obj.channelIndex);
            
            cla(obj.axesHandle);
            hold(obj.axesHandle,'on');
            obj.gObjHandle = plot(obj.axesHandle,obj.streamHandle.timeStamp(obj.timeIndex(t1:t2)),data);
            for it=1:obj.numberOfChannelsToPlot
                set(obj.gObjHandle(it),'color',obj.color(it,:),'userData',{obj.streamHandle, obj.channelIndex(it)});
            end
            %- topoplot on ButtonDown event
            % set(obj.gObjHandle,'ButtonDownFcn',@gObject_ButtonDownFcn);
            resFolder = [fileparts(which('EEGBrowser.m')) filesep 'resources'];
            imgScaleDown = imread([resFolder filesep 'Gnome-list-remove.svg.png']);
            imgScaleUp = imread([resFolder filesep 'Gnome-list-add.svg.png']);
            helpIcon = imread([resFolder filesep 'Gnome-help-browser.svg.png']);
            hDown = uicontrol('Parent', obj.figureHandle, 'Style', 'pushbutton','TooltipString','Scale down','Position',[543+150 53 40 40],'Callback',@onScaleDown,'CData',imgScaleDown);
            hUp = uicontrol('Parent', obj.figureHandle, 'Style', 'pushbutton','TooltipString','Scale up','Position',[543+190 53 40 40],'Callback',@onScaleUp,'CData',imgScaleUp);
            hHelp = uicontrol(obj.figureHandle,'CData',helpIcon,'TooltipString','Help','Position',[543+230 53 40 40],'Callback','web(''https://github.com/aojeda/EEGBrowser#eegbrowser'',''-browser'')');
            set([hDown, hUp hHelp],'Units','Normalized')
            set(obj.figureHandle,'KeyPressFcn',@onKeyPress);
            
            obj.plotThisTimeStamp(obj.nowCursor);
        end
        %%
        function plotThisTimeStamp(obj,nowCursor)
            delta = obj.windowWidth/2;
            if  nowCursor + delta < obj.streamHandle.timeStamp(obj.timeIndex(end)) &&...
                    nowCursor - delta > obj.streamHandle.timeStamp(obj.timeIndex(1))
                newNowCursor = nowCursor;
            elseif nowCursor + delta >= obj.streamHandle.timeStamp(obj.timeIndex(end))
                newNowCursor = obj.streamHandle.timeStamp(obj.timeIndex(end)) - delta;
                if strcmp(get(obj.timerObj,'Running'),'on')
                    stop(obj.timerObj);
                end
            else
                newNowCursor = obj.streamHandle.timeStamp(obj.timeIndex(1)) + delta;
            end
            nowCursor = newNowCursor;
            
            % find now cursor index
            obj.nowCursor = nowCursor;
            [~,t1] = min(abs(obj.streamHandle.timeStamp(obj.timeIndex) - (obj.nowCursor-obj.windowWidth/2)));  
            [~,t2] = min(abs(obj.streamHandle.timeStamp(obj.timeIndex) - (obj.nowCursor+obj.windowWidth/2)));  
            if t1==t2, return;end
            dt = length(t1:t2)/2;
            data = obj.streamHandle.mmfObj.Data.x(obj.timeIndex(t1:t2),obj.channelIndex);
            if sum(data(~isnan(data)))
                if obj.normalizeFlag
                    [data,~,sigma] = zscore(data);
                    sigma(isnan(sigma)) = 1;
                else
                    [~,mu,sigma] = zscore(data);
                    mu(isnan(mu)) = 0;
                    sigma(isnan(sigma)) = 1;
                    data = data - ones(2*dt,1)*mu;
                end
            else
                sigma = 1;
            end
            sigma(sigma == 0) = 1;
            if obj.numberOfChannelsToPlot > 1
                ytick = (1:obj.numberOfChannelsToPlot)*mean(sigma)/obj.gain;
                data = data + ones(2*dt,1)*fliplr(1:obj.numberOfChannelsToPlot)*mean(sigma)/obj.gain;
                delta = abs(diff(ytick([2 1])));
                lim = [ytick(1) - delta ytick(end) + delta];
            elseif obj.numberOfChannelsToPlot == 1 && obj.streamHandle.numberOfChannels > 1
                data = data/max([data; eps]);
                ytick = mean(data);
                mx = 1.5*max(abs(data));
                lim = [ytick - mx ytick + mx];
            else
                if sum(data(:))
                    data = data./max(abs(data));
                    ytick = min(abs([min(data(:)) max(data(:))]));
                    lim = [ytick-max(abs(data))*1.5 ytick+max(abs(data))*1.5];
                else
                    ytick = 0;
                    lim = [-1 1];
                end
            end
            if sum(data(:)) == 0, ytick = 0; lim = [-1 1];end
            
            if obj.numberOfChannelsToPlot <= 1
                data = {data};
            elseif obj.numberOfChannelsToPlot == 2
                data = num2cell(data,1)';
            else
                data = num2cell(data,[1 obj.numberOfChannelsToPlot])';
            end
            set(obj.gObjHandle,'XData',obj.streamHandle.timeStamp(obj.timeIndex(t1:t2)),{'YData'},data,{'Color'},obj.colorInCell);           
            xlim(obj.axesHandle,obj.streamHandle.timeStamp(obj.timeIndex([t1 t2])));
            ylim(obj.axesHandle,lim);  % static limit
            
            if obj.showEvents
                if length(obj.eventObj.latencyInFrame) ~= size(obj.eventColor,1), obj.initEventColor;end
                [~,loc1,loc2] = intersect(obj.streamHandle.timeStamp(obj.timeIndex(t1:t2)),obj.streamHandle.timeStamp(obj.eventObj.latencyInFrame));
                Nloc1 = length(loc1);
                if ~isempty(loc1)
                    hold(obj.axesHandle,'on');
                    set(obj.figureHandle,'CurrentAxes',obj.axesHandle)
                    ind = obj.timeIndex(t1:t2);
                    linesHandler = line(ones(2,1)*obj.streamHandle.timeStamp(ind(loc1)),(ones(length(loc1),1)*lim)','Parent',obj.axesHandle);
                    textPos = [obj.streamHandle.timeStamp(ind(loc1))-0.5*(obj.streamHandle.timeStamp(ind(loc1))-...
                            obj.streamHandle.timeStamp(ind(loc1)-0));ones(1,Nloc1)*lim(2)*1.01];
                    try delete(obj.textHandle);end %#ok
                    set(linesHandler,{'color'},num2cell(obj.eventColor(loc2,:)',[1 3])');
                    obj.textHandle = zeros(length(loc1),1);
                    for it=1:Nloc1
                        obj.textHandle(it) = text('Position',textPos(:,it),'String',obj.eventObj.label(loc2(it)),'Color',obj.eventColor(loc2(it),:),...
                            'Parent',obj.axesHandle,'FontSize',12,'FontWeight','bold','Rotation',45);
                    end
                    boundaryEvents = cellfun(@isempty,obj.eventObj.label(loc2));
                    set(linesHandler(boundaryEvents),'LineStyle','-.','Color','k','LineWidth',2);
                    set(obj.figureHandle,'CurrentAxes',obj.axesHandle)
                    hold(obj.axesHandle,'off');
                end
            end
            set(obj.axesHandle,'YTick',ytick);
            set(obj.axesHandle,'YTickLabel',obj.yTickLabel);
            set(obj.timeTexttHandle,'String',['Current latency = ' num2str(obj.nowCursor,4) ' sec']);
            set(obj.sliderHandle,'Value',obj.nowCursor);
        end
        %%
        function plotStep(obj,step)
            delta = obj.windowWidth/2;
            if obj.nowCursor+step+delta < obj.streamHandle.timeStamp(obj.timeIndex(end)) &&...
                    obj.nowCursor+step-delta > obj.streamHandle.timeStamp(obj.timeIndex(1))
                newNowCursor = obj.nowCursor+step;
            elseif obj.nowCursor+step+delta > obj.streamHandle.timeStamp(obj.timeIndex(end))
                newNowCursor = obj.streamHandle.timeStamp(obj.timeIndex(end))-delta;
            else
                newNowCursor = obj.streamHandle.timeStamp(obj.timeIndex(1))+delta;
            end
            obj.plotThisTimeStamp(newNowCursor);
        end
        %%
        function changeColormap(obj,newColormap)
            if nargin < 2, newColormap = '';end
            cmaps = dir(fileparts(which('colormap')));
            if strcmpi(newColormap,'eegplot')
                obj.color = ones(obj.numberOfChannelsToPlot,1)*[0 0 0.4];
            elseif any(ismember({cmaps.name},[newColormap '.m']))
                obj.color = eval([newColormap '(' num2str(obj.numberOfChannelsToPlot) ')']);
            else
                warndlg(['Colormap ''' newColormap ''' is not available. We will use ''lines'' instead.'])
                obj.color = lines(obj.numberOfChannelsToPlot);
                newColormap = 'lines';
            end
            obj.colormap = newColormap;
        end
        %%
        function onClose(obj,~,~)
            delete(obj);
        end
        function delete(obj)
            delete(obj.figureHandle);
            delete(obj.streamHandle.mmfName);
        end
        
        %%
        function obj = changeSettings(obj)
            prompt = {'Scale (use +/- keys to change the scale only)','Channels to plot','Speed (values between 1 an 5)','Page width (in seconds)','Normalize','Show channel number','Show events','Colormap (any MATLAB colormap or eegplot for eegplot color)'};
            defaultVal = {num2str(obj.gain), ['[' num2str(obj.channelIndex) ']'], num2str(obj.speed), num2str(obj.windowWidth), num2str(obj.normalizeFlag),...
                num2str(obj.showChannelNumber), num2str(obj.showEvents), obj.colormap};
            properties = inputdlg(prompt,'Preferences',1,defaultVal);
            
            if isempty(properties)
                return;
            end
            obj.gain = abs(str2num(properties{1}));                 %#ok
            obj.channelIndex = str2num(properties{2});              %#ok
            tmp = str2double(properties{3});
            obj.speed = interp1(1:5,1:5,tmp,'nearest','extrap');
            obj.windowWidth = abs(str2num(properties{4}));          %#ok
            obj.normalizeFlag = logical(str2num(properties{5}));    %#ok
            obj.showChannelNumber = logical(str2num(properties{6}));%#ok
            
            showEvents = str2num(properties{7});                    %#ok
            if ~showEvents
                obj.eventObj = event;
            end
            obj.changeColormap(properties{8});
            
            figure(obj.figureHandle);
            obj.init();
        end
    end
    %%
    methods(Static)
        function updateChannelDependencies(~,evnt)
            evnt.AffectedObject.numberOfChannelsToPlot = length(evnt.AffectedObject.channelIndex);
            
            evnt.AffectedObject.yTickLabel = cell(evnt.AffectedObject.numberOfChannelsToPlot,1);
            labels = cell(evnt.AffectedObject.numberOfChannelsToPlot,1);
            channels = evnt.AffectedObject.channelIndex;
            
            if evnt.AffectedObject.showChannelNumber
                if evnt.AffectedObject.numberOfChannelsToPlot > 1
                    for jt=fliplr(1:evnt.AffectedObject.numberOfChannelsToPlot), labels{jt} = num2str(channels(evnt.AffectedObject.numberOfChannelsToPlot-jt+1));end
                else
                    labels{1} = num2str(channels);
                end
            else
                if evnt.AffectedObject.numberOfChannelsToPlot > 1
                    if isempty(evnt.AffectedObject.streamHandle.label)
                        for jt=fliplr(1:evnt.AffectedObject.numberOfChannelsToPlot), labels{jt} = num2str(evnt.AffectedObject.numberOfChannelsToPlot-jt+1);end
                    else
                        for jt=fliplr(1:evnt.AffectedObject.numberOfChannelsToPlot), labels{jt} = evnt.AffectedObject.streamHandle.label{channels(evnt.AffectedObject.numberOfChannelsToPlot-jt+1)};end
                    end
                else
                    if ~isempty(evnt.AffectedObject.streamHandle.label)
                        labels = evnt.AffectedObject.streamHandle.label(evnt.AffectedObject.channelIndex);
                    else
                        labels{1} = '';
                    end
                end
            end
            evnt.AffectedObject.yTickLabel = labels;
            evnt.AffectedObject.changeColormap(evnt.AffectedObject.colormap);
        end
        %%
        function updateTimeIndexDenpendencies(~,evnt)
            if evnt.AffectedObject.timeIndex(1) ~= -1
                evnt.AffectedObject.nowCursor = evnt.AffectedObject.streamHandle.timeStamp(evnt.AffectedObject.timeIndex(1)) + 2.5;
                set(evnt.AffectedObject.sliderHandle,'Min',evnt.AffectedObject.streamHandle.timeStamp(evnt.AffectedObject.timeIndex(1)));
                set(evnt.AffectedObject.sliderHandle,'Max',evnt.AffectedObject.streamHandle.timeStamp(evnt.AffectedObject.timeIndex(end)));
                set(evnt.AffectedObject.sliderHandle,'Value',evnt.AffectedObject.nowCursor);
            end
        end
        %%
        function updateColorInCell(~,evnt)
            evnt.AffectedObject.colorInCell = cell(evnt.AffectedObject.numberOfChannelsToPlot,1);
            for it=1:evnt.AffectedObject.numberOfChannelsToPlot
                evnt.AffectedObject.colorInCell{it} = evnt.AffectedObject.color(it,:);
            end
        end
    end
end

%%
function onKeyPress(src,evnt)
obj = src.UserData;
switch evnt.Key
    case 'leftarrow',  plotStep(obj,-obj.step*obj.speed*2);
    case 'rightarrow', plotStep(obj,obj.step*obj.speed*2);
    case 'subtract'
        obj.gain = obj.gain/2;
        plotStep(obj,0);
    case 'add'
        obj.gain = obj.gain*2;
        plotStep(obj,0);
end
end
%%
function onScaleDown(src,evnt)
obj = src.Parent.UserData;
obj.gain = obj.gain/2;
plotStep(obj,0);
end

function onScaleUp(src,evnt)
obj = src.Parent.UserData;
obj.gain = obj.gain*2;
plotStep(obj,0);
end